@@ -18,6 +18,9 @@ from vipy.batch import Batch
 import hashlib
 import torch.utils.data
 from torch.utils.data import DataLoader, random_split
+import bz2
+import pickle
+import time
 
     
 def disjoint_activities(V, activitylist):    
@@ -145,8 +148,11 @@ class Datasets():
         else:
             return self.reload(src, format)
 
-    def new(self, objlist, id):
-        return self.cache(Dataset(objlist, id=id))
+    def new(self, objlist, id=None):
+        """Create a new dataset, cache it and return Dataset()"""
+        d = Dataset(objlist, id=id)
+        self.cache(d)
+        return d
 
     def cached(self):
         return list(self._datasets.keys())
@@ -187,16 +193,16 @@ class Datasets():
         f = lambda net,v,b=batchsize: net.gpu(list(range(torch.cuda.device_count())), batchsize=b*torch.cuda.device_count())(v, d_category_to_object[v.category()]) if v.category() in d_category_to_object else v
         return self.map(src, f, model=model, dst=dst)
 
-    def instance_mining(self, src, dstdir=None, dst=None, batchsize=1, conf=0.05, iou=0.6, maxhistory=5, smoothing=None, objects=None, mincover=0.6, maxconf=0.2):
+    def instance_mining(self, src, dstdir=None, dst=None, batchsize=1, minconf=0.01, miniou=0.6, maxhistory=5, smoothing=None, objects=None, trackconf=0.05):
         model = pycollector.detection.MultiscaleVideoTracker(batchsize=batchsize)
         dst = dst if dst is not None else '%s_instancemining' % (self.load(src).id())
         dstdir = remkdir(os.path.join(self._indir, dst))
-        f_process = lambda net,v,o=objects,dstdir=dstdir: net.gpu(list(range(torch.cuda.device_count()))).track(v, objects=o, verbose=False, conf=conf, iou=iou, maxhistory=maxhistory, smoothing=smoothing, mincover=mincover, maxconf=maxconf).pkl(os.path.join(dstdir, '%s.pkl' % v.videoid())).print()
+        f_process = lambda net,v,o=objects,dstdir=dstdir: net.gpu(list(range(torch.cuda.device_count()))).track(v, objects=o, verbose=False, minconf=minconf, miniou=miniou, maxhistory=maxhistory, smoothing=smoothing, trackconf=trackconf).pkl(os.path.join(dstdir, '%s.pkl' % v.videoid())).print()
         return self.map(src, f_process, model=model, dst=dst)  
 
     def stabilize_refine_activityclip(self, src, dst, batchsize=1, dt=5, minlength=5, maxsize=512*3):
         from vipy.flow import Flow
-        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize)  # =8 (0046) =20 (0053)
+        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize)  
         f_process = (lambda net,v,dt=dt,f=self._schema,dst=dst,maxsize=maxsize: 
                      [a.saveas(f(dst,a,k)).pkl().print() for (k,a) in enumerate(net(Flow(flowdim=256).stabilize(v, strict=False, residual=True, maxsize=maxsize).saveas(f(dst,v), flush=True).pkl().print(),
                                                                                     proposalconf=5E-2, 
@@ -214,14 +220,14 @@ class Datasets():
         V = self.map(src, dst, f_process, model=model, save=False).dataset(dst)
         V = [v for clips in V for v in clips if (v is not None) and not v.hasattribute('unrefined') and not v.hasattribute('unstabilized')]  # unpack clips, remove videos that failed
         V = [v.activityfilter(lambda a: any([a.hastrack(t) and len(t)>minlength and t.during(a.startframe(), a.endframe()) for t in v.tracklist()])) for v in V]  # get rid of activities without tracks greater than dt
-        return self.new(V, dst).save(dst)
+        return self.new(V, dst)
 
     def refine_activityclip_stabilize(self, src, dst, batchsize=1, dt=3, minlength=5, padwidthfrac=1.0, padheightfrac=0.2):
         """Refine the bounding box, split into clips then stabilzie the clips.  This is more memory efficient for stabilization"""
         assert self.has_dataset(src) and self.isloaded(src)
 
         from vipy.flow import Flow
-        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize)  # =8 (0046) =20 (0053)
+        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize)  
         f_process = (lambda net,v,dt=dt,f=self._schema,dst=dst,padwidthfrac=padwidthfrac,padheightfrac=padheightfrac: 
                      [Flow(flowdim=256).stabilize(a, strict=False, residual=True, padwidthfrac=padwidthfrac, padheightfrac=padheightfrac, maxsize=None).saveas(f(dst,a,k)).pkl().print()
                       for (k,a) in enumerate(net(v,
@@ -240,7 +246,7 @@ class Datasets():
         V = self.map(src, dst, f_process, model=model, save=False).dataset(dst)
         V = [v for clips in V for v in clips if (v is not None) and not v.hasattribute('unrefined') and not v.hasattribute('unstabilized')]  # unpack clips, remove videos that failed
         V = [v.activityfilter(lambda a: any([a.hastrack(t) and len(t)>minlength and t.during(a.startframe(), a.endframe()) for t in v.tracklist()])) if minlength is not None else v for v in V]  # get rid of activities without tracks greater than dt
-        return self.new(V, dst).save(dst)
+        return self.new(V, dst)
 
     def refine_stabilize(self, src, dst, batchsize=1, dt=3, minlength=5, padwidthfrac=1.0, padheightfrac=0.2):
         """Refine the bounding box, split into clips then stabilzie the clips.  This is more memory efficient for stabilization"""
@@ -256,7 +262,7 @@ class Datasets():
         assert self.has_dataset(src) and self.isloaded(src)
 
         from vipy.flow import Flow
-        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize)  # =8 (0046) =20 (0053)
+        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize) 
         f_process = (lambda net,v,dt=dt,f=self._schema,dst=dst,padwidthfrac=padwidthfrac,padheightfrac=padheightfrac: 
                      Flow(flowdim=256).stabilize(net(v,
                                                      proposalconf=5E-2, 
@@ -273,10 +279,10 @@ class Datasets():
 
         V = self.map(src, dst, f_process, model=model, save=False).dataset(dst)
         V = [v for v in V if (v is not None) and not v.hasattribute('unrefined') and not v.hasattribute('unstabilized')]  # remove videos that failed
-        return self.new(V, dst).save(dst)
+        return self.new(V, dst)
     
     def refine_activityclip(self, src, dst, batchsize=1, dt=5, minlength=5):
-        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize)  # =8 (0046) =20 (0053)
+        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize) 
         f_process = (lambda net,v,dt=dt,f=self._schema,dst=dst: 
                      [a.saveas(f(dst,a,k)).pkl().print() for (k,a) in enumerate(net(v,
                                                                                     proposalconf=5E-2, 
@@ -294,12 +300,12 @@ class Datasets():
         V = self.map(src, dst, f_process, model=model, save=False).dataset(dst)
         V = [v for clips in V for v in clips if (v is not None) and not v.hasattribute('unrefined')]  # unpack clips, remove videos that failed
         V = [v.activityfilter(lambda a: any([a.hastrack(t) and len(t)>minlength and t.during(a.startframe(), a.endframe()) for t in v.tracklist()])) for v in V]  # get rid of activities without tracks greater than dt
-        return self.new(V,dst).save(dst)
+        return self.new(V,dst)
 
     def pad_activityclip(self, src, dst, t=2):        
         f_tubelet = lambda v, t=t, f=self._schema, dst=dst: [a.saveas(f(dst,a,k)).print() for (k,a) in enumerate(v.activitymap(lambda x: x.padto(t)).activityclip())]
         V = [a for V in self.map(src, dst, f_tubelet, save=False).dataset(dst) for a in V if a is not None]  # unpack
-        return self.new(V, dst).save(dst)
+        return self.new(V, dst)
 
     def stabilize(self, src, dst):
         from vipy.flow import Flow
@@ -308,12 +314,12 @@ class Datasets():
         return self.map(src, dst, f_stabilize)
 
     def refine(self, src, dst, batchsize=1, dt=3, minlength=5):        
-        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize)  # =8 (0046) =20 (0053)
+        model = pycollector.detection.VideoProposalRefinement(batchsize=batchsize) 
         f = lambda net, v, dt=dt: net(v, proposalconf=5E-2, proposaliou=0.8, miniou=0.2, dt=dt, mincover=0.8, byclass=True, shapeiou=0.7, smoothing='spline', splinefactor=None, strict=True).print()
         V = self.map(src, dst, f, model=model, save=False).dataset(dst)
         V = [v for v in V if (v is not None) and (not v.hasattribute('unrefined'))]  # remove videos that failed refinement
         V = [v.activityfilter(lambda a: any([a.hastrack(t) and len(t)>minlength and t.during(a.startframe(), a.endframe()) for t in v.tracklist()])) for v in V]  # get rid of activities without tracks greater than dt
-        return self.new(V,dst).save(dst)
+        return self.new(V, dst)
 
     def nondegenerate(self, src, dst=None):
         dst = dst if dst is not None else '%s_nondegenerate' % self.load(src).id()
@@ -339,25 +345,6 @@ class Datasets():
         f_tubelet = lambda v, m=mindim, f=f_saveas: v.mindim(m).saveas(f(v)).print()
         return self.map(src, f_tubelet).id('mindim_%d' % mindim)        
 
-    def tohtml(self, src, outfile, mindim=512, title='Visualization', fraction=1.0, display=False, clip=True, activities=True, category=True):
-        """Generate a standalone HTML file containing quicklooks for each annotated activity in dataset, along with some helpful provenance information for where the annotation came from"""
-    
-        assert ishtml(outfile), "Output file must be .html"
-        assert fraction > 0 and fraction <= 1.0, "Fraction must be between [0,1]"
-        
-        import vipy.util  # This should not be necessary, but we get "UnboundLocalError" without it, not sure why..
-        import vipy.batch  # requires pip install vipy[all]
-
-        dataset = self.dataset(src)
-        assert all([isinstance(v, vipy.video.Video) for v in dataset])
-        dataset = [dataset[k] for k in np.random.permutation(range(len(dataset)))[0:int(len(dataset)*fraction)]]
-        
-        quicklist = self._Batch(dataset).map(lambda v: (v.load().quicklook(), v.flush().print()) if v.canload() else (None, None)).result()
-        quicklooks = [imq for (imq, v) in quicklist if imq is not None]  # keep original video for HTML display purposes
-        provenance = [{'clip':str(v), 'activities':str(';'.join([str(a) for a in v.activitylist()])), 'category':v.category()} for (imq, v) in quicklist]
-        (quicklooks, provenance) = zip(*sorted([(q,p) for (q,p) in zip(quicklooks, provenance)], key=lambda x: x[1]['category']))  # sorted in category order
-        return vipy.visualize.tohtml(quicklooks, provenance, title='%s' % title, outfile=outfile, mindim=mindim, display=display)
-
     def activitymontage(self, src, outfile, gridrows=30, gridcols=50, mindim=64, bycategory=False):
         """30x50 activity montage, each 64x64 elements using the output of prepare_dataset"""
         vidlist = self.dataset(src)
@@ -392,14 +379,10 @@ class Datasets():
         return self.new(self.load(src).list(), dst)
 
 
-class Transformer(Datasets):
-    def __repr__(self):
-        return str('<pycollector.dataset.Transformer: "%s">' % self._indir)
-
-
 
 class TorchDataset(torch.utils.data.Dataset):
-    def __init__(self, d, f_transformer):
+    """Converter from a pycollector dataset to a torch dataset"""
+    def __init__(self, f_transformer, d):
         assert isinstance(d, Dataset), "Invalid input"
         self._dataset = d
         self._f_transformer = f_transformer
@@ -408,7 +391,43 @@ class TorchDataset(torch.utils.data.Dataset):
         return self._f_transformer(self._dataset[k])
 
     def __len__(self):
-        return len(self._videolist)
+        return len(self._dataset)
+
+
+class TorchDatadir(torch.utils.data.Dataset):
+    """A torch dataset stored as a directory of .json files each containing a single object for training, and a transformer lambda function to conver this object to a tensor/label for data augmented training"""
+    def __init__(self, f_transformer, jsondir):
+        assert os.path.isdir(jsondir)
+        self._jsonlist = vipy.util.listjson(jsondir)
+        self._f_transformer = f_transformer
+
+    def __getitem__(self, k):
+        return self._f_transformer(vipy.util.load(self._jsonlist[k]))
+
+    def __len__(self):
+        return len(self._jsonlist)
+
+
+class TorchTensordir(torch.utils.data.Dataset):
+    """A torch dataset stored as a directory of .pkl.bz2 files each containing a list of tensor/labels used for data augmented training"""
+    def __init__(self, tensordir):
+        assert os.path.isdir(tensordir)
+        self._dirlist = sorted([s for s in vipy.util.extlist(tensordir, '.pkl.bz2')], key=lambda d: int(vipy.util.filetail(d).rsplit('_', 1)[1].split('.pkl.bz2')[0]))
+
+    def __getitem__(self, k):
+        assert k >= 0 and k < len(self._dirlist)
+        for j in range(0,2):
+            try:
+                obj = vipy.util.bz2pkl(self._dirlist[k])
+                assert len(obj) > 0, "Invalid augmentation"
+                return obj[np.random.randint(0, len(obj))]  # choose one tensor at random
+            except:
+                time.sleep(1)  # try again after a bit if another process is augmenting this .pkl.bz2 in parallel
+        print('ERROR: %s corrupted' % self._dirlist[k])
+        return self.__getitem__(np.random.randint(0, len(self)))  # maximum retries reached, get another one
+
+    def __len__(self):
+        return len(self._dirlist)
 
     
 class Dataset():
@@ -549,6 +568,10 @@ class Dataset():
 
     def class_to_index(self):
         return {v:k for (k,v) in enumerate(self.classlist())}
+
+    def index_to_class(self):
+        return {v:k for (k,v) in self.class_to_index().items()}
+
     def label_to_index(self):
         return self.class_to_index()
 
@@ -581,13 +604,32 @@ class Dataset():
         self._objlist = [v for v in self._objlist if f(v)]
         return self
 
-    def take(self, n, category=None):
+    def take_per_category(self, n, id=None):
+        C = vipy.util.groupbyasdict(self._objlist, lambda v: v.category())
+        return Dataset([C[k][j] for (k,v) in C.items() for j in np.random.permutation(list(range(len(v))))[0:n]], id=id)
+
+    def tojsondir(self, outdir):
+        print('[pycollector.dataset]: exporting %d json files to "%s"...' % (len(self), outdir))
+        vipy.util.remkdir(outdir)  # to avoid race condition
+        Batch(vipy.util.chunklist([(k,v) for (k,v) in enumerate(self._objlist)], 64), as_completed=True, minscatter=1).map(lambda X: [vipy.util.save(x[1].clone(), os.path.join(outdir, '%s_%d.json' % (x[1].clone().videoid(), x[0]))) for x in X]).result()
+        return outdir
+
+    def takelist(self, n, category=None):
         objlist = self._objlist if category is None else [v for v in self._objlist if v.category() == category]
-        return [objlist[k] for k in random.sample(range(0, len(objlist)), n)]
+        assert n <= len(objlist), "Invalid length"
+        return [objlist[k] for k in random.sample(range(0, len(objlist)), n)]  # without replacement
+
+    def take(self, n, category=None):
+        return Dataset(self.takelist(n, category))
     
-    def split(self, trainfraction=0.8, testfraction=0.1, valfraction=0.1, seed=42):
+    def split(self, trainfraction=0.9, valfraction=0.1, testfraction=0, seed=42):
         """Split the dataset by category by fraction so that video IDs are never in the same set"""
         assert self.isvipy(), "Invalid input"
+        assert trainfraction >=0 and trainfraction <= 1
+        assert valfraction >=0 and valfraction <= 1
+        assert testfraction >=0 and testfraction <= 1
+        assert trainfraction + valfraction + testfraction == 1.0
+
         np.random.seed(seed)
         A = self.list()
         
@@ -595,14 +637,17 @@ class Dataset():
         videoid = list(set([a.videoid() for a in A]))
         np.random.shuffle(videoid)
         (testid, valid, trainid) = vipy.util.dividelist(videoid, (testfraction, valfraction, trainfraction))        
+        (testid, valid, trainid) = (set(testid), set(valid), set(trainid))
         d = vipy.util.groupbyasdict(A, lambda a: 'testset' if a.videoid() in testid else 'valset' if a.videoid() in valid else 'trainset')
-        (trainset, testset, valset) = (d['trainset'], d['testset'], d['valset'])        
+        (trainset, testset, valset) = (d['trainset'] if 'trainset' in d else [], 
+                                       d['testset'] if 'testset' in d else [], 
+                                       d['valset'] if 'valset' in d else [])
 
         print('[pycollector.dataset]: trainset=%d (%1.1f)' % (len(trainset), trainfraction))
         print('[pycollector.dataset]: valset=%d (%1.1f)' % (len(valset), valfraction))
         print('[pycollector.dataset]: testset=%d (%1.1f)' % (len(testset), testfraction))
         
-        return (Dataset(trainset, id='trainset'), Dataset(valset, id='valset'), Dataset(testset, id='testset'))
+        return (Dataset(trainset, id='trainset'), Dataset(valset, id='valset'), Dataset(testset, id='testset') if len(testset)>0 else None)
 
     def tocsv(self, csvfile=None):
         csv = [v.csv() for v in self.list]        
@@ -624,9 +669,110 @@ class Dataset():
         return self
     
     def count(self):
+        """Counts for each label"""
         assert self.isvipy()
         return vipy.util.countby(self.list(), lambda v: v.category())
 
+    def percentage(self):
+        """Fraction of dataset for each label"""
+        d = self.count()
+        n = sum(d.values())
+        return {k:v/float(n) for (k,v) in d.items()}
+
+    def multilabel_inverse_frequency_weight(self):
+        fw = {k:0 for k in self.classlist()}
+        for v in self.list():
+            lbl_frequency = vipy.util.countby([a for A in v.activitylabel() for a in A], lambda x: x)  # frequency within clip
+            lbl_weight = {k:(f/float(max(1, sum(lbl_frequency.values())))) for (k,f) in lbl_frequency.items()}  # multi-label likelihood within clip, sums to one
+            for (k,w) in lbl_weight.items():
+                fw[k] += w
+
+        n = sum(fw.values())  
+        ifw = {k:1.0/((w/n)*len(fw)) for (k,w) in fw.items()}
+        return ifw
+
+    def inverse_frequency_weight(self):
+        d = {k:1.0/max(v,1) for (k,v) in self.count().items()}
+        n = sum(d.values())
+        return {k:len(d)*(v/float(n)) for (k,v) in d.items()}
+
+    def collectors(self, outfile=None):
+        assert self.isvipy()
+        d = vipy.util.countby(self.list(), lambda v: v.attributes['collector_id'])
+        f = lambda x,n: len([k for (k,v) in d.items() if int(v) >= n])
+        print('[collector.dataset.collectors]: Collectors = %d ' % f(d,0))
+        print('[collector.dataset.collectors]: Collectors with >10 submissions = %d' % f(d,10))
+        print('[collector.dataset.collectors]: Collectors with >100 submissions = %d' % f(d,100))
+        print('[collector.dataset.collectors]: Collectors with >1000 submissions = %d' % f(d,1000))
+        print('[collector.dataset.collectors]: Collectors with >10000 submissions = %d' % f(d,10000))
+
+        if outfile is not None:
+            from vipy.metrics import histogram
+            histogram(d.values(), list(range(len(d.keys()))), outfile=outfile, ylabel='Submissions', xlabel='Collector ID', xrot='vertical', fontsize=3, xshow=False)            
+        return d
+
+    def os(self, outfile=None):
+        assert self.isvipy()
+        d = vipy.util.countby([v for v in self.list() if v.hasattribute('device_identifier')], lambda v: v.attributes['device_identifier'])
+        print('[collector.dataset.collectors]: Device OS = %d ' % len(d))
+        if outfile is not None:
+            from vipy.metrics import pie
+            pie(d.values(), d.keys(), explode=None, outfile=outfile,  shadow=False)
+        return d
+
+    def device(self, outfile=None, n=24, fontsize=7):
+        d_all = vipy.util.countby([v for v in self.list() if v.hasattribute('device_type') and v.attributes['device_type'] != 'unrecognized'], lambda v: v.attributes['device_type'])
+        
+        topk = [k for (k,v) in sorted(list(d_all.items()), key=lambda x: x[1])[-n:]] 
+        other = np.sum([v for (k,v) in d_all.items() if k not in set(topk)])
+
+        d = {k:v for (k,v) in d_all.items() if k in set(topk)}
+        d.update( {'Other':other} )
+        d = dict(sorted(list(d.items()), key=lambda x: x[1]))
+
+        print('[collector.dataset.collectors]: Device types = %d ' % len(d_all))
+        print('[collector.dataset.collectors]: Top-%d Device types = %s ' % (n, str(topk)))
+
+        if outfile is not None:
+            from vipy.metrics import pie
+            pie(d.values(), d.keys(), explode=None, outfile=outfile,  shadow=False, legend=False, fontsize=fontsize, rotatelabels=False)
+        return d
+        
+    def duration_in_frames(self, outfile=None):
+        assert self.isvipy()
+        d = {k:np.mean([v[1] for v in v]) for (k,v) in vipy.util.groupbyasdict([(a.category(), len(a)) for v in self.list() for a in v.activitylist()], lambda x: x[0]).items()}
+        if outfile is not None:
+            from vipy.metrics import histogram
+            histogram(d.values(), d.keys(), outfile=outfile, ylabel='Duration (frames)', fontsize=6)            
+        return d
+
+    def duration_in_seconds(self, outfile=None):
+        assert self.isvipy()
+        d = {k:np.mean([v[1] for v in v]) for (k,v) in vipy.util.groupbyasdict([(a.category(), len(a)/v.framerate()) for v in self.list() for a in v.activitylist()], lambda x: x[0]).items()}
+        if outfile is not None:
+            from vipy.metrics import histogram
+            histogram(d.values(), d.keys(), outfile=outfile, ylabel='Duration (seconds)', fontsize=6)            
+        return d
+
+    def framerate(self, outfile=None):
+        assert self.isvipy()
+        d = vipy.util.countby([int(round(v.framerate())) for v in self.list()], lambda x: x)
+        if outfile is not None:
+            from vipy.metrics import pie
+            pie(d.values(), ['%d fps' % k for k in d.keys()], explode=None, outfile=outfile,  shadow=False)
+        return d
+        
+        
+    def density(self, outfile=None):
+        assert self.isvipy()
+        d = [len(v) for (k,v) in vipy.util.groupbyasdict(self.list(), lambda v: v.videoid()).items()]
+        d = vipy.util.countby(d, lambda x: x)
+        if outfile is not None:
+            from vipy.metrics import histogram
+            histogram(d.values(), d.keys(), outfile=outfile, ylabel='Frequency', xlabel='Activities per video', fontsize=6, xrot=None)            
+        return d
+        
+
     def stats(self, outdir=None, object_categories=['Person', 'Car'], plot=True):
         """Analyze the dataset to return helpful statistics and plots"""
         assert self.isvipy()
@@ -711,7 +857,32 @@ class Dataset():
         return d
 
     def totorch(self, f_video_to_tensor):
-        return TorchDataset(self.clone(), f_video_to_tensor)
+        return TorchDataset(f_video_to_tensor, self)
+
+    def totorch_datadir(self, f_video_to_tensor, outdir):
+        return TorchDatadir(f_video_to_tensor, self.tojsondir(outdir))
+
+    def tohtml(self, outfile, mindim=512, title='Visualization', fraction=1.0, display=False, clip=True, activities=True, category=True):
+        """Generate a standalone HTML file containing quicklooks for each annotated activity in dataset, along with some helpful provenance information for where the annotation came from"""
+    
+        assert ishtml(outfile), "Output file must be .html"
+        assert fraction > 0 and fraction <= 1.0, "Fraction must be between [0,1]"
+        
+        import vipy.util  # This should not be necessary, but we get "UnboundLocalError" without it, not sure why..
+        import vipy.batch  # requires pip install vipy[all]
+
+        dataset = self.list()
+        assert all([isinstance(v, vipy.video.Video) for v in dataset])
+        dataset = [dataset[k] for k in np.random.permutation(range(len(dataset)))[0:int(len(dataset)*fraction)]]
+        dataset = [v for v in dataset if all([len(a) < 15*v.framerate() for a in v.activitylist()])]  # remove extremely long videos
+
+        quicklist = vipy.batch.Batch(dataset, strict=False, as_completed=True, minscatter=1).map(lambda v: (v.load().quicklook(), v.flush().print())).result()
+        quicklist = [x for x in quicklist if x is not None]  # remove errors
+        quicklooks = [imq for (imq, v) in quicklist]  # keep original video for HTML display purposes
+        provenance = [{'clip':str(v), 'activities':str(';'.join([str(a) for a in v.activitylist()])), 'category':v.category()} for (imq, v) in quicklist]
+        (quicklooks, provenance) = zip(*sorted([(q,p) for (q,p) in zip(quicklooks, provenance)], key=lambda x: x[1]['category']))  # sorted in category order
+        return vipy.visualize.tohtml(quicklooks, provenance, title='%s' % title, outfile=outfile, mindim=mindim, display=display)
+
 
     
 class ActivityDataset(Dataset):
